#ifndef __JABYENGINE_PRIMITIVE_RECTANGLE_TYPES_HPP__
#define __JABYENGINE_PRIMITIVE_RECTANGLE_TYPES_HPP__
#include "primitive_support_types.hpp"

namespace JabyEngine {
    namespace GPU {
        namespace internal {
            struct RectCode : public CodeBase<RectCode> {
                enum struct Size : uint8_t {
                    Variable    = 0,
                    Pixel1x1    = 1,
                    Sprite8x8   = 2,
                    Sprite16x16 = 3
                };
                static constexpr uint8_t CmdValue = 0b011;

                static constexpr auto Size = BitRange::from_to(27 - BitCorrection, 28 - BitCorrection);
            };

            template<typename T>
            struct RectCodeInterface {
                typedef typename ::JabyEngine::GPU::internal::RectCode Code;

                constexpr T& set_semi_transparent(bool set = true) {
                    static_cast<T*>(this)->code.set_tenary(set, Code::SemiTransparent, Code::NonTransparent);
                    return *static_cast<T*>(this);
                }
            };

            template<typename RectCode::Size Size> 
            struct RECT_F : public RectCodeInterface<RECT_F<Size>> {
                typedef RECT_F<Size>::Code Code;
                static constexpr auto IdentityCode = Code().set(Code::Size.with(static_cast<uint8_t>(Size))).set(Code::Untextured).set(Code::NonTransparent);

                Color24 color;
                Code    code;
                Vertex  position;   

                constexpr RECT_F() = default;
                constexpr RECT_F(const Color24& color, const Vertex& position) : color(color), code(IdentityCode), position(position) {
                }
            };

            template<typename RectCode::Size Size> 
            struct RECT_T : public RectCodeInterface<RECT_T<Size>> {
                typedef RECT_T<Size>::Code Code;
                static constexpr auto IdentityCode = Code(RECT_F<Size>::IdentityCode).set(Code::Textured);

                Color24      color;
                Code         code;
                Vertex       position;
                PagePosition page;

                constexpr RECT_T() = default;
                constexpr RECT_T(const Color24& color, const Vertex& position, const PagePosition& page) : color(color), code(IdentityCode), position(position), page(page) {
                }
            };
        }

        typedef internal::RECT_F<internal::RectCode::Size::Pixel1x1>    RECT_F1;
        typedef internal::RECT_F<internal::RectCode::Size::Sprite8x8>   RECT_F8;
        typedef internal::RECT_F<internal::RectCode::Size::Sprite16x16> RECT_F16;
        struct RECT_FVAR : public internal::RECT_F<internal::RectCode::Size::Variable> {
            Vertex position_bottom_right;

            constexpr RECT_FVAR() = default;
            constexpr RECT_FVAR(const Color24& color, const AreaI16& area) : RECT_F(color, area.position), position_bottom_right(area.get_bottom_left()) {
            }
        };

        typedef internal::RECT_T<internal::RectCode::Size::Pixel1x1>    RECT_T1;
        typedef internal::RECT_T<internal::RectCode::Size::Sprite8x8>   RECT_T8;
        typedef internal::RECT_T<internal::RectCode::Size::Sprite16x16> RECT_T16;
        struct RECT_TVAR : public internal::RECT_T<internal::RectCode::Size::Variable> {
            Vertex position_bottom_right;

            constexpr RECT_TVAR() = default;
            constexpr RECT_TVAR(const Color24& color, const AreaI16& area, const PagePosition& page) : RECT_T(color, area.position, page), position_bottom_right(area.get_bottom_left()) {
            }
        };

        namespace internal {
            __jaby_engine_declare_render_primitive(RECT_F1);
            __jaby_engine_declare_render_primitive(RECT_F8);
            __jaby_engine_declare_render_primitive(RECT_F16);
            __jaby_engine_declare_render_primitive(RECT_FVAR);

            __jaby_engine_declare_render_primitive(RECT_T1);
            __jaby_engine_declare_render_primitive(RECT_T8);
            __jaby_engine_declare_render_primitive(RECT_T16);
            __jaby_engine_declare_render_primitive(RECT_TVAR);
        }
    }
}

#endif //!__JABYENGINE_PRIMITIVE_RECTANGLE_TYPES_HPP__